__all__ = ["is_none", "normalise_to_string_and_uppercase", "list"]

from . import list

from ... import constants

def is_none(thread: str) -> bool:
    return True if thread is None or thread == "NONE" else False # NB: Has to be string and upppercase.

def normalise_to_string_and_uppercase(thread: str) -> str: # The thread list iterator only supports strings and numbers and not None, hence the renaming to "NONE".
    return constants.none_value() if is_none(thread) else str(thread).upper()



from ... import error
from ...model.thread_item import ThreadItem
from ...model.timer import TimerObject

def try_get_thread_item_and_index(timer: TimerObject, thread: str) -> tuple[ThreadItem, int]:
    try:
        index_counter = 0
        for thread_item in timer.threads:
            if thread_item.name == thread: # Check if thread already exists in list. Note that this expects a normalised thread (i.e. string and not None) and already in uppercase so it evaluates as intended.
                return thread_item, index_counter # Make sure only the first match is returned, yet the main function is not designed to allow input of duplicates.
            else:
                index_counter += 1
        return None, None
    except Exception:
        error.message_for_action(f"in the Timer's lookup module", thread = thread)

def add(timer: TimerObject, thread: str, start_time: int, decimals: int) -> None:
    try:
        timer.threads.append(ThreadItem(name = thread, start_time = start_time, decimals = decimals))
    except Exception:
        error.message_for_action(f"when trying to add entry to the Timer's thread list", thread = thread)

def remove(timer: TimerObject, entry_index: int) -> None:
    try:
        timer.threads.pop(entry_index)
    except Exception:
        error.message_for_action(f"when trying to remove entry from the Timer's thread list for entry index \"{entry_index}\"")



from .. import constants
from . import colour
from .. import error
from ..model.timer import TimerObject

def mediate(timer: TimerObject, decimals: int) -> int: # If the start function doesn't have decimals defined, then use the decimals value defined when the Timer() was initiated.
    return timer.decimals if decimals is None else validate_and_normalise(decimals)

def validate_and_normalise(decimals: int) -> int:
    try:
        if isinstance(decimals, str) is True or decimals is None:
            print(f"{colour.yellow()}Timer: Decimals set to default {constants.decimals.default()} due to invalid input.{colour.reset()}")
            return constants.decimals.default()
        elif decimals > 9:
            print(f"{colour.yellow()}Timer: Decimals set to 9 as the Timer doesn't support more than 9 decimals (i.e. nanoseconds).{colour.reset()}")
            return 9
        elif decimals in range(0, 10):
            return int(decimals)
        else:
            print(f"{colour.yellow()}Timer: Decimals set to default {constants.decimals.default()} due to invalid input.{colour.reset()}")
            return constants.decimals.default()
    except Exception:
        error.message_for_action(f"when trying to verify the Timer's decimals input \"{decimals}\"")



from ..model.elapsed_time_fractions import ElapsedTimeFractions

def calculate_time_fractions(elapsed_time_ns: int) -> ElapsedTimeFractions: # Elapsed time is in nanoseconds and should be calculated as difference between start and stop time using on the time.perf_counter_ns() function. 
	microseconds, nanoseconds = divmod(elapsed_time_ns, 1000)
	milliseconds, microseconds = divmod(microseconds, 1000) if microseconds > 0 else (0, 0) # As divmod() can be slow, let's return 0s as a tuple if divmod() isn't needed.
	seconds, milliseconds = divmod(milliseconds, 1000) if milliseconds > 0 else (0, 0)
	minutes, seconds = divmod(seconds, 60) if seconds > 0 else (0, 0)
	hours, minutes = divmod(minutes, 60) if minutes > 0 else (0, 0)
	days, hours = divmod(hours, 24) if hours > 0 else (0, 0)

	return ElapsedTimeFractions(
		nanoseconds = int(nanoseconds), 
		microseconds = int(microseconds),
		milliseconds = int(milliseconds),
		seconds = int(seconds),
		minutes = int(minutes),
		hours = int(hours),
		days = int(days))

class TimeFractions:
	def __init__(self, elapsed_time_ns: int) -> None:
		self.time: ElapsedTimeFractions = calculate_time_fractions(elapsed_time_ns)

	def count_microseconds_to_float(self) -> float:
		return self.time.microseconds + self.time.nanoseconds / 1000
		
	def count_milliseconds_to_float(self) -> float:
		return self.time.milliseconds + self.count_microseconds_to_float() / 1000 # This could potentially be faster by dividing self.microseconds by a 1000 directly, yet we don't want to lose precision in the decimals.

	def count_seconds_to_float(self) -> float:
		return self.time.seconds + self.count_milliseconds_to_float() / 1000 # This could potentially be faster by dividing self.milliseconds by a 1000 directly, yet we don't want to lose precision in the decimals.

	def seconds_rounded(self) -> float: # If 2 seconds and 567 milliseconds, ensure it'll be rounded up to 3 seconds.
		return int(round(self.count_seconds_to_float(), 0))

	def count_minutes_to_seconds(self) -> float:
		return self.time.minutes * 60 + self.count_seconds_to_float()



import time
from . import constants
from . import controller
from . import error
from . import helper
from .model.thread_item import ThreadItem
from .model.timer import TimerObject

class Timer(TimerObject):
	_instance = None

	def __new__(cls, *args, **kwargs) -> TimerObject:
		if not cls._instance: # Singleton: Ensure there's only a single instance of Timer running.
			cls._instance = TimerObject.__new__(cls, *args, **kwargs)
		return cls._instance

	def __init__(self, decimals: int = constants.decimals.default()) -> None:
		self.threads: list[ThreadItem] = []
		self.decimals: int = decimals if decimals == constants.decimals.default() else helper.decimals.validate_and_normalise(decimals)

	def start(self, thread: str = None, decimals: int = None) -> None:
		try:
			start_time = time.perf_counter_ns() # For precision, this is the first operation of the function.
			controller.start(self, thread, start_time, decimals)
		except Exception:
			error.message_for_action("when trying to start the Timer", thread = thread)

	def stop(self, thread: str = None) -> None:
		try:
			stop_time = time.perf_counter_ns() # For precision, this is the first operation of the function.
			controller.stop(self, thread, stop_time)
		except Exception:
			error.message_for_action("when trying to stop the Timer", thread = thread)



__all__ = ["start_controller", "stop_controller", "message_for_action"]

from .. import helper
from ..helper import colour
from ..model.timer import TimerObject

def start_controller(thread: str) -> None:
    if helper.thread.is_none(thread):
        print(f"{colour.yellow()}Timer is running. Use .stop() to stop it.{colour.reset()}")
    else:
        print(f"{colour.yellow()}Timer for thread {thread} is running. Use .stop({thread = }) to stop it.{colour.reset()}")

def stop_controller(timer: TimerObject, thread: str) -> None:
    if helper.thread.is_none(thread):
        print(f"{colour.yellow()}Timer is not running. Use .start() to start it.{colour.reset()}")
    else:
        print(f"{colour.yellow()}Timer for thread {thread} is not running. Use .start({thread = }) to start it.{colour.reset()}")
    if len(timer.threads) > 0:
        open_threads = [thread_item.name for thread_item in timer.threads]
        print(f"Or maybe you aren't stopping the right thread? Currently open threads: {', '.join(open_threads)}")

def message_for_action(action: str, thread: str = None) -> None:
    print(f"{colour.yellow()}Timer: Something went wrong {action}{'' if helper.thread.is_none(thread) else f' for thread {thread}'}.{colour.reset()}")



from .. import error
from . import colour
from .. import helper
from ..helper.time_fractions import TimeFractions

def message(thread: str, fractions: TimeFractions, decimals: int) -> None:
    try:
        text_intro = f"Elapsed time{'' if helper.thread.is_none(thread) else f' (thread {colour.green()}{thread}{colour.reset()})'}:"
        time = fractions.time
        if time.days > 0:
            print(f"{text_intro} {time.days}d {time.hours}h {time.minutes}m {fractions.seconds_rounded()}s") # Format: 1d 2h 3m 4s
        elif time.hours > 0:
            print(f"{text_intro} {time.hours}h {time.minutes}m {fractions.seconds_rounded()}s") # Format: 1h 2m 3s
        elif time.minutes > 0:
            print(f"{text_intro} {fractions.count_minutes_to_seconds():.{decimals}f} seconds ({time.minutes}m {fractions.seconds_rounded()}s)") # Format: 62.34 seconds (1m 2s)
        elif time.seconds > 0:
            print(f"{text_intro} {fractions.count_seconds_to_float():.{decimals}f} seconds") # Format: 0.123456789 seconds
        elif time.milliseconds > 0:
            print(f"{text_intro} {fractions.count_milliseconds_to_float():.{decimals}f} milliseconds") # Format: 123.45 milliseconds
        elif time.microseconds > 0:
            print(f"{text_intro} {fractions.count_microseconds_to_float():.{decimals}f} microseconds") # Format: 234.56 microseconds
        else:
            print(f"{text_intro} {time.nanoseconds} nanoseconds") # Format: 345 nanoseconds
    except Exception:
        error.message_for_action(f"in the Timer's output message module", thread = thread)



from dataclasses import dataclass

@dataclass(frozen = True)
class ThreadItem:
    """Class to define a thread's name, start time, and decimals."""

    name: str
    start_time: str
    decimals: int



from abc import ABC, abstractmethod

class TimerObject(ABC):
    """Abstract base class of Timer object."""

    @abstractmethod
    def start(self, thread: str, decimals: int) -> None:
        """Starts the Timer"""
    
    @abstractmethod
    def stop(self, thread: str) -> None:
        """Stops the Timer"""
